import lxml.html

from latestos.scraper.mirrors import ArizonaMirror

class ArchScraper(ArizonaMirror):
    """ Latest Arch Version Checker """
    URL = "https://mirror.arizona.edu/archlinux/iso/latest/"
    OS_NAME = "Arch"

    def get_latest_release(self) -> tuple:
        """
        Gets data from mirror and then parses it (html).

        Returns:
            (HtmlElement, str): parsed html data, release url
        """
        r = self.fetch(self.__class__.URL)

        # Parse the HTML
        document = lxml.html.fromstring(r)

        return document, self.__class__.URL

    def link_is_for_iso_file(self, href: str) -> bool:
        """
        Checks whether a link (href) corresponds to a release iso file

        Returns:
            (bool): is iso file
        """
        return href.endswith(".iso")

    def link_is_for_iso_checksum(self, href: str) -> bool:
        """
        Checks whether a link (href) corresponds to a release iso checksum file

        Returns:
            (bool): is iso checksum file
        """
        return href == "sha1sums.txt"
    


import requests

import lxml.html

VAGRANTUP_BOX_URL = "https://app.vagrantup.com/ProfessorManhattan/"

VAGRANTUP_BOX_OS_LIST = [
    "ubuntu",
    "arch"
]


def vagrantup_check(os_name: str, version: str) -> bool:
    """
    Checks whether an OS is on version :version on the vagrantup box.

    Returns:
        (bool): is OS updated?
    """
    # Fetch url & parse HTML
    r = requests.get(VAGRANTUP_BOX_URL)
    document = lxml.html.fromstring(r.text)

    # Get list of boxes titles, each in the ":os-name :version" format
    boxes = document.xpath(".//h4")

    # Go through each box
    for box in boxes:
        # Format box title
        formatd_box = " ".join(box.xpath("string()").replace("\n", "").split())

        # Extract OS name & version
        box_os_name, box_os_version = get_os_data_from_box_title(formatd_box)

        # if current box corresponds to the os we're looking for, check if
        # os version is up to date
        if os_name in box_os_name.lower():
            if version in box_os_version:
                # Do a final check for the version
                v_nums = box_os_version.replace(version, "").replace(".", "")

                if not v_nums or (v_nums.isnumeric() and int(v_nums) == 0):
                    return True

    return False


def get_os_data_from_box_title(box_title: str) -> tuple:
    """
    Gets OS name and version from a properly formatted box title.

    Returns:
        (str, str): OS name, OS version
    """
    # Split box title into [os name, os version]
    box_title_elements = box_title.split(" v")

    return box_title_elements[0], box_title_elements[1]



from latestos.scraper.mirrors import ArizonaMirror


class RaspbianScraper(ArizonaMirror):
    """
    Latest Raspbian Version Checker
    Note that this does not extract data from an Arizona Mirror
    """
    URL = "https://downloads.raspberrypi.org/rpd_x86/images/"
    OS_NAME = "Raspbian"

    def get_latest_release_url(self, mirror_links: list) -> str:
        """
        Check all mirror links and return the one that corresponds to the
        latest release.

        Returns:
            (str): latest release url
        """
        return self.parse_latest_release_url(mirror_links[-1])

    def link_is_for_iso_file(self, href: str) -> bool:
        """
        Checks whether a link (href) corresponds to a release iso file

        Returns:
            (bool): is iso file
        """
        return href.endswith(".iso")

    def link_is_for_iso_checksum(self, href: str) -> bool:
        """
        Checks whether a link (href) corresponds to a release iso checksum file

        Returns:
            (bool): is iso checksum file
        """
        return href.endswith(".sha1")

    def get_filename_sections(self, iso_filename: str) -> list:
        """
        Extracts filename data from the iso_filename.
        By default it splits by "-".

        Returns:
            (list): filename data - second element is the os version
        """
        sections = iso_filename.split("-raspios", 1)

        # Replace version dashes - to dots .
        version = sections[0].replace("-", ".")

        # Remember to place ISO version on second element
        return [sections[1], version]



from latestos.scraper.mirrors import ArizonaMirror


class FedoraScraper(ArizonaMirror):
    """ Latest Fedora Version Checker """
    URL = "https://mirror.arizona.edu/fedora/linux/releases/"
    RELEASE_ISO_URL_SUFFIX = "Server/x86_64/iso/"
    OS_NAME = "Fedora"

    def get_latest_release_url(self, mirror_links: list) -> str:
        """
        Check all mirror links and return the one that corresponds to the
        latest release.

        Returns:
            (str): latest release url
        """
        latest_release_number, latest_release_url = 0, None

        # Go through mirror links
        for link in mirror_links:
            # Check if link is an OS version link (only numbers)
            if link.text[:-1].isnumeric():
                # Extract number
                release_link_number = float(link.text[:-1])

                # Update latest release if necessary
                if release_link_number > latest_release_number:
                    latest_release_number = release_link_number
                    latest_release_url = self.parse_latest_release_url(link)

        return latest_release_url

    def link_is_for_iso_file(self, href: str) -> bool:
        """
        Checks whether a link (href) corresponds to a release iso file

        Returns:
            (bool): is iso file
        """
        return href.startswith("Fedora-Server-dvd") and href.endswith(".iso")

    def link_is_for_iso_checksum(self, href: str) -> bool:
        """
        Checks whether a link (href) corresponds to a release iso checksum file

        Returns:
            (bool): is iso checksum file
        """
        return href.endswith("CHECKSUM")

    def get_filename_sections(self, iso_filename: str) -> list:
        """
        Extracts filename data from the iso_filename.

        Returns:
            (list): filename data - second element is the os version
        """
        # Split filename
        arch = "x86_64-"
        sections = iso_filename.split(arch)

        if len(sections) >= 1:
            # Format the version before returning it
            version = sections[1].replace(".iso", "").replace("-", ".")
            sections = [sections[0], version]

        return sections



import sys
import subprocess

from latestos.scraper.utils import get_os_scraper
from latestos.files.json import update_json_release_file

from latestos.vagrantup.check import VAGRANTUP_BOX_OS_LIST, vagrantup_check


DEFAULT_JSON_FILENAME = "./template.json"


def get_latest_os():
    # Get os_name, json_filename, bash_command from command line arguments
    os_name, json_filename, bash_command = get_params()
    run(os_name, json_filename, bash_command)


def run(
    os_name: str,
    json_filename: str = DEFAULT_JSON_FILENAME,
    bash_command: list = []
) -> None:
    """ Entry point for the script """
    # Get the scraper depending on the OS name
    scraper = get_os_scraper(os_name)

    # Get latest release data
    iso_url, checksum_url, version = scraper.get_latest_release_data()

    # Update the JSON file
    update_json_release_file(json_filename, iso_url, checksum_url, version)
    print(f"Updated {json_filename}")

    # if OS is on vagrantup box os list, check if the version is up to date
    if bash_command and os_name in VAGRANTUP_BOX_OS_LIST:
        uptodate = vagrantup_check(os_name, version)

        if not uptodate:
            run_subprocess(bash_command)


def get_params() -> tuple:
    """
    Get parameters from command line 

    Returns:
        (str, str, list): os name, json filename, bash command
    """
    # If there is only one argument (script name), raise an exception
    if len(sys.argv) <= 1:
        raise ValueError(
            "You need to pass the OS name " \
            "(ubuntu, centos, fedora, debian, raspbian or arch)"
        )

    args = sys.argv[1:]

    # Extract the necessary variables
    os_name = args[0]
    json_filename = args[1] if len(args) > 1 else DEFAULT_JSON_FILENAME
    bash_command = args[2:] if len(args) > 2 else []

    return os_name, json_filename, bash_command


def run_subprocess(bash_command: list, verbose: bool = True) -> None:
    """
    Runs bash command.
    """
    # Open subprocess and run
    proc = subprocess.Popen(bash_command, stdout=subprocess.PIPE, shell=True)
    output, _ = proc.communicate()

    # Print the output if necessary
    if verbose:
        print(output.decode("utf-8"))


if __name__ == "__main__":
    get_latest_os()



from latestos.scraper.mirrors import ArizonaMirror


class UbuntuScraper(ArizonaMirror):
    """ Latest Ubuntu Version Checker """
    URL = "https://mirror.arizona.edu/ubuntu-releases/"
    OS_NAME = "Ubuntu"

    def get_latest_release_url(self, mirror_links: list) -> str:
        """
        Check all mirror links and return the one that corresponds to the
        latest release.

        Returns:
            (str): latest release url
        """
        latest_release_url = None

        # Go through mirror links
        for link in mirror_links:
            # Parse link
            link_version = link.text
            dotless_link = link_version[:-1].replace(".", "")

            # If link is an OS version link (only numbers), update release url
            if dotless_link.isnumeric():
                latest_release_url = self.parse_latest_release_url(link)

        return latest_release_url

    def link_is_for_iso_file(self, href: str) -> bool:
        """
        Checks whether a link (href) corresponds to a release iso file

        Returns:
            (bool): is iso file
        """
        return href.endswith("server-amd64.iso")

    def link_is_for_iso_checksum(self, href: str) -> bool:
        """
        Checks whether a link (href) corresponds to a release iso checksum file

        Returns:
            (bool): is iso checksum file
        """
        return href == "SHA256SUMS"



from abc import ABC

import requests
from lxml.html import HtmlElement


class BaseScraper(ABC):
    """
    BaseScraper is an abstract class, which means it can not be used by itself.
    It defines common functionality amongst all latest OS scrapers.
    One must override all non implemented methods and the URL variable.
    """
    # Root URL of the mirror
    URL = None

    def fetch(self, url: str = None) -> str:
        """
        Fetches data from an url, defaults to the mirror's root URL.

        Returns:
            (str): response text
        """
        if url:
            r = requests.get(url)
        else:
            r = requests.get(self.__class__.URL)
        return r.text

    def get_latest_release_data(self) -> tuple:
        """
        Gets latest OS release data.

        Returns:
            (str, str, str): iso url, checksum url, os version
        """
        # Get latest release
        release, release_url = self.get_latest_release()

        # Extract data from latest release
        filename, iso_url = self.get_iso_filename_and_url(release, release_url)
        checksum_url = self.get_iso_checksum_url(release, filename, release_url)
        version = self.get_iso_version(filename)

        return iso_url, checksum_url, version

    def get_latest_release(self) -> tuple:
        """
        Find latest release.

        Returns:
            (HtmlElement, str): release html data, release url
        """
        raise NotImplementedError()

    def get_iso_filename_and_url(self, release: HtmlElement, release_url: str) -> tuple:
        """
        Extracts ISO filename and URL from release.

        Returns:
            (str, str): iso filename, iso url
        """
        raise NotImplementedError()

    def get_iso_checksum_url(self, release: HtmlElement, iso_filename: str, release_url: str) -> str:
        """
        Extracts ISO checksum URL from release.

        Returns:
            str: checksum url
        """
        raise NotImplementedError()

    def get_iso_version(self, iso_filename: str, release_url: str) -> str:
        """
        Extracts ISO version from release.

        Returns:
            str: iso version
        """
        raise NotImplementedError()



from latestos.scraper.mirrors import ArizonaMirror


class CentOSScraper(ArizonaMirror):
    """ Latest CentOS Version Checker """
    URL = "https://mirror.arizona.edu/centos/"
    RELEASE_ISO_URL_SUFFIX = "isos/x86_64/"
    OS_NAME = "CentOS"

    def get_latest_release_url(self, mirror_links: list) -> str:
        """
        Check all mirror links and return the one that corresponds to the
        latest release.

        Returns:
            (str): latest release url
        """
        latest_release_number, latest_release_url = 0, None

        # Go through mirror links
        for link in mirror_links:
            link_version = link.text

            # Check if "stream"
            if "stream" not in link_version:
                continue

            version_number = link_version.replace("-stream", "") \
                                         .replace(".", "")

            # Check if link is an OS version link (only numbers)
            if version_number.isnumeric():
                # Extract number
                release_link_number = float(version_number[:-1])

                # Update latest release if necessary
                if release_link_number > latest_release_number:
                    latest_release_number = release_link_number
                    latest_release_url = self.parse_latest_release_url(link)

        return latest_release_url

    def link_is_for_iso_file(self, href: str) -> bool:
        """
        Checks whether a link (href) corresponds to a release iso file

        Returns:
            (bool): is iso file
        """
        return href.endswith("-dvd1.iso")

    def link_is_for_iso_checksum(self, href: str) -> bool:
        """
        Checks whether a link (href) corresponds to a release iso checksum file

        Returns:
            (bool): is iso checksum file
        """
        return href.endswith("CHECKSUM")



import json


def update_json_release_file(
    json_file: str, iso_url: str, checksum_url: str, version: str):
    """
    Reads a JSON file and updates the latest OS release variables.
    """
    # Read JSON
    data = read_json_file(json_file)

    # Create the "variables" field if it doesn't exist
    has_variables_field = data.get("variables", None) != None
    if not has_variables_field:
        data["variables"] = {}

    # Update the latest OS variables
    data["variables"]["iso_url"] = iso_url
    data["variables"]["iso_checksum_url"] = checksum_url
    data["variables"]["iso_version"] = version

    # Save changes
    write_json_file(data, json_file)


def read_json_file(json_file: str) -> dict:
    """
    Reads JSON file.

    Returns:
        (dict): JSON data
    """
    try:
        # Try to read the JSON data from the file
        with open(json_file) as f:
            return json.loads(f.read())
    except FileNotFoundError:
        # If the file does not exist, use the default template
        with open("./latestos/templates/template.json") as f:
            return json.loads(f.read())
    except Exception:
        # If there's an unexpected error, then just re-raise the exception
        raise Exception(f"Could not read {json_file}")


def write_json_file(data, json_file: str):
    """
    Saves JSON in a file.
    """
    try:
        # Try to open the file and write the json to it
        with open(json_file, "w", encoding="utf-8") as f:
            json.dump(data, f, indent=4)
    except Exception:
        raise Exception(f"Could not write new data to {json_file}")



from latestos.scraper.mirrors import ArizonaMirror

import lxml.html
from lxml.html import HtmlElement


class DebianScraper(ArizonaMirror):
    """ Latest Debian Version Checker """
    URL = "https://mirror.arizona.edu/debian-cd/current-live/amd64/iso-hybrid/"
    SHASUMS_URL = "https://mirror.arizona.edu/debian-cd/current-live/amd64/iso-hybrid/SHA1SUMS"
    OS_NAME = "Debian"

    def get_latest_release(self) -> tuple:
        """
        Gets data from mirror and then parses it (html).

        Returns:
            (HtmlElement, str): parsed html data, release url
        """
        r = self.fetch(self.__class__.URL)

        # Parse the HTML
        document = lxml.html.fromstring(r)

        return document, self.__class__.URL

    def link_is_for_iso_file(self, href: str) -> bool:
        """
        Checks whether a link (href) corresponds to a release iso file

        Returns:
            (bool): is iso file
        """
        return href.endswith("amd64-standard.iso")

    def get_iso_checksum_url(self, release: HtmlElement, iso_filename: str, release_url: str) -> str:
        """
        Extracts ISO checksum URL from release.

        Returns:
            str: checksum url
        """
        return self.__class__.SHASUMS_URL

    def get_filename_sections(self, iso_filename: str) -> list:
        """
        Extracts filename data from the iso_filename.
        By default it splits by "-".

        Returns:
            (list): filename data - second element is the os version
        """
        # Ignore first element so that ISO version goes second and not third
        return iso_filename.split("-")[1:]



from latestos.scraper.base import BaseScraper
from latestos.scraper.arch import ArchScraper
from latestos.scraper.centos import CentOSScraper
from latestos.scraper.fedora import FedoraScraper
from latestos.scraper.ubuntu import UbuntuScraper
from latestos.scraper.debian import DebianScraper
from latestos.scraper.raspbian import RaspbianScraper


def get_os_scraper(os_name: str) -> BaseScraper:
    """
    Takes an OS name and returns the corresponding scraper.

    Returns:
        (BaseScraper): scraper
    """
    os_name = os_name.lower()

    if os_name == "arch":
        return ArchScraper()
    elif os_name == "centos":
        return CentOSScraper()
    elif os_name == "fedora":
        return FedoraScraper()
    elif os_name == "ubuntu":
        return UbuntuScraper()
    elif os_name == "debian":
        return DebianScraper()
    elif os_name == "raspbian":
        return RaspbianScraper()
    else:
        raise ValueError(f"{os_name}")



from setuptools import setup, find_packages


setup(
    name='latestos',
    packages=find_packages(),
    version='0.4.1',
    license='MIT',
    description='Latest OS version checker for Linux Distros using the Arizona Mirror',
    author='Renny Montero',
    author_email='rennym19@gmail.com',
    url='https://github.com/rennym19/latestos',
    download_url='https://github.com/rennym19/latestos/archive/v_04_01.tar.gz',
    keywords=['OS', 'LINUX', 'VERSION', 'CHECKER', 'SCRAPER'],
    install_requires=[
        'requests',
        'lxml',
    ],
    include_package_data=True,
    entry_points={'console_scripts': ['latestos = latestos.commands:get_latest_os']},
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
)



from abc import ABC

import lxml.html
from lxml.html import HtmlElement

from latestos.scraper.base import BaseScraper


class ArizonaMirror(BaseScraper, ABC):
    """
    Defines common functions for latest OS scrapers using the Arizona Mirror.
    ArizonaMirror is an abstract class like BaseScraper, but it has a lot of
    functionality already defined.
    """
    URL = ""
    RELEASE_ISO_URL_SUFFIX = ""
    OS_NAME = "OS"

    def get_latest_release(self) -> tuple:
        """
        Find latest release.

        Returns:
            (HtmlElement, str): release html data, release url
        """
        # Fetch the Mirror URL
        r = self.fetch(self.__class__.URL)

        # Parse the HTML and gather all links
        document = lxml.html.fromstring(r)
        links = document.xpath(".//a")

        # Get the link that corresponds to the latest release
        latest_release_url = self.get_latest_release_url(links)

        # If the link was not found, raise error
        if not latest_release_url:
            raise Exception(
                f"Could not find latest {self.__class__.OS_NAME} release")

        # Fetch data for latest release, parse the html, then return both
        r = self.fetch(latest_release_url)
        return lxml.html.fromstring(r), latest_release_url

    def get_latest_release_url(self, mirror_links: list) -> str:
        """
        Check all mirror links and return the one that corresponds to the
        latest release.

        Returns:
            (str): latest release url
        """
        raise NotImplementedError()

    def parse_latest_release_url(self, link: HtmlElement) -> str:
        """
        Takes a link HTML element and parses the URL from it

        Returns:
            (str): parsed and formatted latest release url
        """
        url = link.xpath(".//@href")[0]
        return f'{self.__class__.URL}{url}{self.__class__.RELEASE_ISO_URL_SUFFIX}'

    def get_iso_filename_and_url(self, release: HtmlElement, release_url: str) -> tuple:
        """
        Extracts ISO filename and URL from release.

        Returns:
            (str, str): iso filename, iso url
        """
        # Get the files in the current mirror folder
        files = release.xpath(".//a")

        # Go through the files until we find the iso file
        for f in files:
            href = f.xpath("./@href")
            href = href[0] if len(href) else None

            if href and self.link_is_for_iso_file(href):
                return f.text, f"{release_url}{href}"

        raise Exception("Could not find ISO filename and URL")

    def link_is_for_iso_file(self, href: str) -> bool:
        """
        Checks whether a link (href) corresponds to a release iso file

        Returns:
            (bool): is iso file
        """
        raise NotImplementedError()

    def get_iso_checksum_url(self, release: HtmlElement, iso_filename: str, release_url: str) -> str:
        """
        Extracts ISO checksum URL from release.

        Returns:
            str: checksum url
        """
        # Get the files in the current mirror folder
        files = release.xpath(".//a")

        # Go through the files until we find the iso checksum file
        for f in files:
            href = f.xpath("./@href")
            href = href[0] if len(href) else None

            if href and self.link_is_for_iso_checksum(href):
                return f"{release_url}{href}"

        raise Exception(f"Could not find {iso_filename} checksum")

    def link_is_for_iso_checksum(self, href: str) -> bool:
        """
        Checks whether a link (href) corresponds to a release iso checksum file

        Returns:
            (bool): is iso checksum file
        """
        raise NotImplementedError()

    def get_iso_version(self, iso_filename: str) -> str:
        """
        Extracts ISO version from release.

        Returns:
            str: iso version
        """
        filename_sections = self.get_filename_sections(iso_filename)

        # If the filename was properly extracted, return it
        if len(filename_sections) >= 2:
            return filename_sections[1]

        raise Exception(f"Could not extract {iso_filename} OS version")

    def get_filename_sections(self, iso_filename: str) -> list:
        """
        Extracts filename data from the iso_filename.
        By default it splits by "-".

        Returns:
            (list): filename data - second element is the os version
        """
        return iso_filename.split("-")
